/**
 * W-SLDA Toolkit
 * Engine version: 2022.06.11
 * API version: 20220221
 * */

/**
 * W-SLDA Toolkit
 * Engine version: 2022.02.21
 * */


/** 
 * EXTERNAL POTENTIAL V_ext
 * @param ix x-coordinate from range [0,NX), to convert to Cartesian use: x = DX*(ix-NX/2)
 * @param iy y-coordinate from range [0,NY), to convert to Cartesian use: y = DY*(iy-NY/2),
 *           NOTE: in case of 1d code iy=0
 * @param iz z-coordinate from range [0,NZ), to convert to Cartesian use: z = DZ*(iz-NZ/2)
 *           NOTE: in case of 1d and 2d codes iz=0
 * @param it time iteration, use dc_t0 + dc_dt*it to compute corresponding time 
 * @param spin spin indicator, value from set {SPINA,SPINB}
 * @param params array of input parameters, before call of this routine the params array is processed by process_params() routine
 * @param extra_data_size size of extra_data in bytes, if extra_data size=0 the optional data is not uploaded
 * @param extra_data optional set of data uploaded by load_extra_data()
 * @return value of the external potential V_spin(x,y,z)
 * */
__device__ double v_ext(int ix, int iy, int iz, int it, int spin, double *params, size_t extra_data_size, void *extra_data)
{
//     double x = DX*(ix-NX/2);
//     double y = DY*(iy-NY/2);     // for 1d code iy will be always 0
//     double z = DZ*(iz-NZ/2);     // for 1d and 2d codes iz will be always 0
//     double t = dc_t0 + dc_dt*it; // time

    // ADD HERE FORMULA FOR V_ext(r)
    double V_ext = 0.0;

    return V_ext; 
}
 
/** 
 * EXTERNAL PAIRING POTENTIAL Delta_ext
 * @param ix x-coordinate from range [0,NX), to convert to Cartesian use: x = DX*(ix-NX/2)
 * @param iy y-coordinate from range [0,NY), to convert to Cartesian use: y = DY*(iy-NY/2)
 *           NOTE: in case of 1d code iy=0
 * @param iz z-coordinate from range [0,NZ), to convert to Cartesian use: z = DZ*(iz-NZ/2)
 *           NOTE: in case of 1d and 2d codes iz=0
 * @param it time iteration, use dc_t0 + dc_dt*it to compute corresponding time 
 * @param delta - value of delta computed self-consistently for given iteration it. 
 * @param params array of input parameters, before call of this routine the params array is processed by process_params() routine
 * @param extra_data_size size of extra_data in bytes, if extra_data size=0 the optional data is not uploaded
 * @param extra_data optional set of data uploaded by load_extra_data()
 * @return value of external pairing potential Delta_{ext}(x,y,z)
 * Complex type is equivalent to thrust::complex<double>
 * for more info see: https://thrust.github.io/doc/structthrust_1_1complex.html
 * */
__device__ Complex delta_ext(int ix, int iy, int iz, int it, Complex delta, double *params, size_t extra_data_size, void *extra_data)
{
     double x = DX*(ix-NX/2);
     double y = DY*(iy-NY/2);     // for 1d code iy will be always 0
     double z = DZ*(iz-NZ/2);     // for 1d and 2d codes iz will be always 0
     double t = dc_t0 + dc_dt*it; // time

    // ADD HERE FORMULA FOR Delta_ext(r)
     //    Complex D_ext = Complex(0.0,0.0);
     double A = params[20]*exp(-0.5*pow((t-params[21])/params[22], 2));
     double u = 0.0;
     u = A * sin(2.*M_PI*x/params[23]) * sin(2.*M_PI*y/params[24]) * sin(2.*M_PI*z/params[25]);


     //return noise_delta_ext(ix, iy, iz, it, delta);
     return Complex(u,0.0); 
}

//__device__ Complex noise_delta_ext(int ix, int iy, int iz, int it, Complex delta)
//{
  //double _x = (double)(ix)*DX - 1.0*(LX/2);
  //double _y = (double)(iy)*DY - 1.0*(LY/2);
  //double t = dc_t0 + dc_dt*it;
  //double x = DX*(ix-NX/2);
  //double y = DY*(iy-NY/2);
  //double z = DZ*(iz-NZ/2);


  //double A = params[20]*exp(-0.5*pow((t-params[21])/params[22], 2));
  //double u=0.0;
  //if(dc_params[30]>0.5)
  //u = A * sin(2.*M_PI*x/params[23]) * sin(2.*M_PI*y/params[23]); // 2D case
  //else               u = A * sin(2.*M_PI*_x/dc_params[5]); // 1D case


  //return Complex(u,0.0);
  //}


/** 
 * EXTERNAL VELOCITY FIELD vec[v]_ext
 * @param ix x-coordinate from range [0,NX), to convert to Cartesian use: x = DX*(ix-NX/2)
 * @param iy y-coordinate from range [0,NY), to convert to Cartesian use: y = DY*(iy-NY/2)
 *           NOTE: in case of 1d code iy=0
 * @param iz z-coordinate from range [0,NZ), to convert to Cartesian use: z = DZ*(iz-NZ/2)
 *           NOTE: in case of 1d and 2d codes iz=0
 * @param it time iteration, use dc_t0 + dc_dt*it to compute corresponding time 
 * @param spin spin indicator, value from set {SPINA,SPINB}
 * @param coordinate - Cartesian coordinate of the external velocity vector that should be computed, value from set {XAXIS, YAXIS, ZAXIS}
 *                     NOTE: for 1d code only XAXIS is requested, for 2d code XAXIS and YAXIS are requested.
 * @param params array of input parameters, before call of this routine the params array is processed by process_params() routine
 * @param extra_data_size size of extra_data in bytes, if extra_data size=0 the optional data is not uploaded
 * @param extra_data optional set of data uploaded by load_extra_data()
 * @return value of the external velocity vector v_ext(x,y,z)
 * */
__device__ double velocity_ext(int ix, int iy, int iz, int it, int spin, int coordinate, double *params, size_t extra_data_size, void *extra_data)
{
//     double x = DX*(ix-NX/2);
//     double y = DY*(iy-NY/2);     // for 1d code iy will be always 0
//     double z = DZ*(iz-NZ/2);     // for 1d and 2d codes iz will be always 0
//     double t = dc_t0 + dc_dt*it; // time

    // ADD HERE FORMULAS FOR vec{v}_ext=(vx, vy, vz)
    double v_ext;
    if(coordinate==XAXIS) v_ext=0.0;
    if(coordinate==YAXIS) v_ext=0.0;
    if(coordinate==ZAXIS) v_ext=0.0;

    return v_ext; 
}

/** 
 * THIS FUNCTION IS CALLED AT THE BEGINNING OF SIMULATION.
 * After loading params array from input file, the parameters are processed by this routine.
 * @param params array of size MAX_USER_PARAMS with parameters from input file. 
 * @param kF typical Fermi momentum scale of the problem. 
 *           kF=referencekF if the referencekF tag is indicated in the input file, 
 *           otherwise to kF value is assigned according formula kF=(3*pi^2*n)^{1/3}, where n corresponds to density in the box center.
 *           Note that it is passed via a pointer, to access/modify it use kF[0] or (*kF).
 * @param mu array with chemical potentials: mu[SPINA] and mu[SPINB]. 
 * @param extra_data_size size of extra_data in bytes, if extra_data size=0 the optional data is not uploaded
 * @param extra_data optional set of data uploaded by load_extra_data()
 * */
extern "C" void process_params(double *params, double *kF, double *mu, size_t extra_data_size, void *extra_data)
{
    // PROCESS INPUT FILE PARAMETERS 
  params[1] *= kF[0] * kF[0] / 2.00;
  params[2] *= kF[0] * kF[0] / 2.00;
  
  params[21]/=kF[0] * kF[0] / 2.00;
  
  params[23] = LX / params[23]; // in the input I provide how may waves I want to have
  params[24] = LY / params[24]; // in the input I provide how may waves I want to have
  params[25] = LZ / params[25]; // in the input I provide how may waves I want to have
  
  
}

/**
 * THIS FUNCTION IS CALLED AFTER EACH COMPUTATION OF POTENTIALS
 * IT IS CALLED ONLY IF ENABLE_MODIFY_POTENTIALS IS DEFINED (in predefines.h)
 * Before each diagonalization process, user can modify arbitrarily potentials
 * @param it iteration number
 * @param h_densities structure with densities, see (wiki) documentation for list of fields
 *                    NOTE: densities structure is processed by modify_densities(...) function before call ot this function.
 * @param h_potentials struture with potentials, see (wiki) documentation for list of fields
 * Global variabls deliver by tdwslda_functionals_framework_enable.h are:
 * @param params array of input parameters, before call of this routine the params array is processed by process_params() routine
 * @param extra_data_size size of extra_data in bytes, if extra_data size=0 the optional data is not uploaded
 * @param extra_data optional set of data uploaded by load_extra_data()
 * */
#include "tdwslda_functionals_framework_enable.h" // DO NOT REMOVE!
__global__ void modify_potentials(int it, wslda_density h_densities, wslda_potential h_potentials)
{    
    size_t ixyz= threadIdx.x + blockIdx.x * blockDim.x; // compute for this point
    int ix, iy, iz, i;
    
    if(ixyz<NUMBER_ELEMENT)
    {
      //h_potentials.delta[ixyz]=h_potentials.delta[ixyz]*params20*exp(-0.5*pow((t-params21)/params22,2))*sin(k_x*ix)*sin(k_y*iy);

    // decode_ixyz2ixiyiz(ixyz,ix,iy,iz, i);
        // // Now ix, iy, iz keeps lattice coordinate. 
        
        // h_potentials.V_a[ixyz] stores value of spin-up particles mean-field potential for coordinate (x,y,z)
        // and similarly for other potentials
        // ... below you can modify them your wish ...
        // 
    }
}
#include "tdwslda_functionals_framework_disable.h" // DO NOT REMOVE!

/**
 * This function provides size of extra_data array, in bytes.
 * The extra_data of specified size will be allocated by the main process.
 * This function is thread-safe.
 * @param params with input file parameters. 
 *              NOTE: the array contains bare input file values, not processed by process_params()!
 * @return size of the extra_data array that needs to be allocated, if 0 then extra_data will not be allocated.
 * */
extern "C" size_t get_extra_data_size(double *params)
{
    return 0;
}

/**
 * This function loads data into extra_data array.
 * This function is thread-safe.
 * @param size size of array computed using function get_extra_data_size()
 * @param extra_data pointer to array that should be filled with data
 * @param params with input file parameters. 
 *               NOTE: the array contains bare input file values, not processed by process_params()!
 * @return 0 if load is successful, otherwise return error code. If nonzero value is returned the main code will terminate.
 * */
extern "C" int load_extra_data(size_t size, void *extra_data, double *params)
{
    return 0;
}

/**
 * Scattering length, in code units.
 * This function is meaningful only in the case of BDG or SLDAE functionals.
 * For SLDA and ASLDA the scattering length is assumed to be infinite, and the function is ignored.
 * @param ix x-coordinate from range [0,NX), to convert to Cartesian use: x = DX*(ix-NX/2)
 * @param iy y-coordinate from range [0,NY), to convert to Cartesian use: y = DY*(iy-NY/2),
 *           NOTE: in case of 1d code iy=0
 * @param iz z-coordinate from range [0,NZ), to convert to Cartesian use: z = DZ*(iz-NZ/2)
 *           NOTE: in case of 1d and 2d codes iz=0
 * @param it time iteration, use dc_t0 + dc_dt*it to compute corresponding time 
 * @param params array of input parameters, before call of this routine the params array is processed by process_params() routine
 * @param extra_data_size size of extra_data in bytes, if extra_data size=0 the optional data is not uploaded
 * @param extra_data optional set of data uploaded by load_extra_data()
 * @return value of the scattering length a(x,y,z,t).
 * */
__device__ double scattering_length(int ix, int iy, int iz, int it, double *params, size_t extra_data_size, void *extra_data)
{
  // return dc_sclgth; // by default return value from input file.
  double t = dc_t0 + dc_dt*it; // time

  double t0 = params[0];

  double omega1 = params[1];
  double omega2 = params[2];

  double t1 = t0 + M_PI / omega1;
  double t2 = t1 + M_PI / omega2;

  double sclgth0 = params[10];
  double sclgth1 = params[11];
  double sclgth2 = params[12];

    if (t <= t0) {
      return sclgth0;
    } else if (t > t0 && t <= t1) {
      return sclgth0 + (sclgth1-sclgth0) / 2.00 * (1.00 - cos(omega1 * (t-t0)));
    } else if (t > t1 && t <= t2) {
      return sclgth1 + (sclgth2-sclgth1) / 2.00 * (1.00 - cos(omega2 * (t-t1)));;
    } else {
      return sclgth2;
    }

    // however, here you can define your own prescription
    // it can be time and position dependent
    // ...
}

/**
 * ------------------------ FOR FUNCTIONAL == CUSTOMEDF ------------------------
 * @see hpc-engine/tdwslda_functionals.h for more details
 * */

#if FUNCTIONAL==CUSTOMEDF
#include "tdwslda_functionals_framework_enable.h" // DO NOT REMOVE!
__global__ void tdwslda_compute_potentials(int it, wslda_density h_densities, wslda_potential h_potentials, double cccoeff)
{
    size_t ixyz= threadIdx.x + blockIdx.x * blockDim.x; // compute for this point
    int ix, iy, iz, i;
    
//     double na,nb;
//     double Va_ext,Vb_ext;
    
    if(ixyz<NUMBER_ELEMENT)
    {
        decode_ixyz2ixiyiz(ixyz,ix,iy,iz, i); // decode cartesian coordinates
        
//         // external potential
//         Va_ext=u_ext(ix,iy,iz,it,SPINA);
//         Vb_ext=u_ext(ix,iy,iz,it,SPINB);
//         
//         // read densities
//         na=h_densities.rho_a[ixyz];
//         nb=h_densities.rho_b[ixyz];
        
//         // save potential to global memory
//         h_potentials.V_a[ixyz]=...;  // <-- mean field + external potential
//         h_potentials.V_b[ixyz]=...;  // <-- mean field + external potential 
//         h_potentials.delta[ixyz]=...;  // <-- mean field + external potential
//         ...
    }
}


__global__ void tdwslda_compute_energy(int it, wslda_density h_densities, wslda_potential h_potentials, // <-- INPUT
                                       double *E_kin, double *E_pot, double *E_pair, double *E_curr)    // <- OUTPUT
{
    size_t ixyz= threadIdx.x + blockIdx.x * blockDim.x; // compute for this point
    
    double na, nb;
    double taua, taub;
    double jxa, jya, jza, jxb, jyb, jzb;
    
    if(ixyz<NUMBER_ELEMENT)
    {
        // densities
        na=h_densities.rho_a[ixyz];
        nb=h_densities.rho_b[ixyz];
        
        // kinetic energy
        taua=h_densities.tau_a[ixyz]; 
        taub=h_densities.tau_b[ixyz];  
        
        // currents
#if CODEDIM>=1
        jxa=h_densities.j_a_x[ixyz];
        jxb=h_densities.j_b_x[ixyz];
#endif
#if CODEDIM>=2
        jya=h_densities.j_a_y[ixyz];
        jyb=h_densities.j_b_y[ixyz];
#else        
        jya=0.0;
        jyb=0.0;
#endif
#if CODEDIM>=3
        jza=h_densities.j_a_z[ixyz];
        jzb=h_densities.j_b_z[ixyz];
#else
        jza=0.0;
        jzb=0.0;
#endif
        
//         taua-=p_regularization(na)*(jxa*jxa+jya*jya+jza*jza)/na; // -ja^2/na: correction for tilde{tau}_a
//         taub-=p_regularization(nb)*(jxb*jxb+jyb*jyb+jzb*jzb)/nb; // -jb^2/nb: correction for tilde{tau}_b
// 
//         // galilean invariant contribution
//         E_kin[ixyz]=0.5*(h_potentials.alpha_a[ixyz]*taua + h_potentials.alpha_b[ixyz]*taub)*VOLUME_ELEMENT;
//         
//         // potential energy
//         E_pot[ixyz]=0.0;
//         
//         // pairing energy
//         E_pair[ixyz]=(h_potentials.delta[ixyz]*thrust::conj(h_densities.nu[ixyz])).real()*(-1.0)*VOLUME_ELEMENT;
//         
//         // flow energy
//         E_curr[ixyz]= p_regularization(na)*(jxa*jxa + jya*jya + jza*jza)/(2.*na)*VOLUME_ELEMENT  
//                     + p_regularization(nb)*(jxb*jxb + jyb*jyb + jzb*jzb)/(2.*nb)*VOLUME_ELEMENT;

    }
}
#include "tdwslda_functionals_framework_disable.h" // DO NOT REMOVE!
#endif
